<?php
/**
 * @author      Lefteris Kavadas
 * @copyright   Copyright (c) 2016 - 2017 Lefteris Kavadas / firecoders.com
 * @license     GNU General Public License version 3 or later
 */
defined('_JEXEC') or die;

class plgSystemRoute66 extends JPlugin
{
    protected static $language;

    public function onAfterInitialise()
    {
        $application = JFactory::getApplication();
        if ($application->isSite() && $application->getCfg('sef')) {

            // Set language
            self::setLanguage();

            // Initialize our router
            include_once JPATH_SITE.'/plugins/system/route66/lib/router.php';
            $Route66 = new Route66SefRouter();

            // Attach router callbacks
            $router = $application->getRouter();
            $router->attachBuildRule(array($Route66, 'preBuildRoute'), 'preprocess');
            $router->attachBuildRule(array($Route66, 'postBuildRoute'), 'postprocess');
            $router->attachParseRule(array($Route66, 'preParseRoute'), 'preprocess');
            $router->attachParseRule(array($Route66, 'postParseRoute'), 'postprocess');
        }
    }

    public function onAfterDispatch()
    {
        $application = JFactory::getApplication();
        if ($application->isSite() && $application->getCfg('sef')) {

            // Get document
            $document = JFactory::getDocument();

            $componentParams = JComponentHelper::getParams('com_route66');
            $this->params->set('redirect', $componentParams->get('redirect', 1));
            $this->params->set('canonical', $componentParams->get('canonical', 1));
            $this->params->set('exclusions', $componentParams->get('exclusions', array()));

            // Get option
            $option = $application->input->getCmd('option');

            // Duplicate URLs handling
            if ($application->input->getMethod() == 'GET' && ($this->params->get('canonical', 1) || $this->params->get('redirect', 1)) && $document->getType() == 'html' && !in_array($option, $this->params->get('exclusions', array()))) {

                // Get router
                $router = $application->getRouter();

                // Get URI
                $uri = JUri::getInstance();

                // Merge all the available vars
                $vars = array_merge($router->getVars(), $uri->getQuery(true));
                $vars = array_filter($vars);

                // Canonical URI
                $url = $uri->getScheme().'://'.$uri->getHost();
                $port = $uri->getPort();
                if($port) {
                  $url .= ':'.$port;
                }
                $url .= JRoute::_('index.php?'.http_build_query($vars), false);
                $canonical = JUri::getInstance($url);
                $query = $canonical->getQuery(true);
                $canonical = $canonical->toString(array('scheme', 'host', 'port', 'path'));

                // Current URI
                $current = $uri->toString(array('scheme', 'host', 'port', 'path'));

                // Canonical and current URI should match. If not redirect
                if ($this->params->get('redirect', 1) && $current != JUri::root() && urldecode($current) != urldecode($canonical)) {

                    // Redirect
                    $redirect = $canonical;
                    if(count($query)) {
                      $redirect .= '?'.http_build_query($query);
                    }
                    $application->redirect($redirect, 301);
                }

                // Add a canonical link
                if ($this->params->get('canonical', 1)) {
                    // Exclude paginated pages
                    if (!isset($query['start'])) {
                        $document->addHeadLink(htmlspecialchars($canonical), 'canonical');
                    }
                }
            }
        }
    }

    public function onBeforeCompileHead()
    {
        $application = JFactory::getApplication();
        $document = JFactory::getDocument();
        if ($application->isSite() && $document->getType() == 'html') {
            $params = JComponentHelper::getParams('com_route66');
            if ($params->get('facebookPageId')) {
                $document->addCustomTag('<meta property="fb:pages" content="'.$params->get('facebookPageId').'" />');
            }
        }
    }

    protected static function setLanguage()
    {
        // Initialize object with default values
        $language = new stdClass();
        $language->code = '*';
        $language->sef = '';

        // Get application
        $application = JFactory::getApplication();

        // Proceed only if language filter is enabled
        if ($application->isSite() && $application->getLanguageFilter()) {

            // Get default site language
            $default = JComponentHelper::getParams('com_languages')->get('site', 'en-GB');

            // Get avaliable languages by SEF key
            $languages = JLanguageHelper::getLanguages('sef');

            // Do not rely on JFactory::getLanguage()->getTag() because it has not set yet if there is a language change in front-end
            $path = substr(JUri::current(), strlen(JUri::root(false)));
            if (strpos($path, 'index.php') === 0) {
                $path = substr($path, 9);
            }
            if (strlen($path) > 1 && strpos($path, '/') === 0) {
                $path = substr($path, 1);
            }
            $parts = explode('/', $path);
            if ($parts[0] && isset($languages[$parts[0]])) {
                $current = $languages[$parts[0]]->lang_code;
            }

            // If no language is detected fallback to the site default
            if (!isset($current)) {
                $current = $default;
            }
            $language->code = $current;

            // Check for the option "Remove URL Language Code" of the "System - Language Filter" to find out if the core router will add the language prefix
            $languageFilterPlugin = JPluginHelper::getPlugin('system', 'languagefilter');
            $languageFilterPluginParams = new JRegistry($languageFilterPlugin->params);

            $languageFilterPluginPrefixStatus = (bool) ($current != $default || !$languageFilterPluginParams->get('remove_default_prefix'));
            if ($languageFilterPluginPrefixStatus) {
                $languages = JLanguageHelper::getLanguages('lang_code');
                if (isset($languages[$language->code])) {
                    $language->sef = $languages[$language->code]->sef;
                }
            }
        }

        self::$language = $language;
    }

    public static function getLanguage()
    {
        return self::$language;
    }

    public function onAfterRender()
    {
        $this->optimize();
        return true;
    }

    protected function optimize() {

      $application = JFactory::getApplication();
      $user = JFactory::getUser();
      $document = JFactory::getDocument();

      $componentParams = JComponentHelper::getParams('com_route66');
      $this->params->set('optimizeJs', $componentParams->get('optimizeJs', 0));
      $this->params->set('optimizeJsMethod', $componentParams->get('optimizeJsMethod', 'file'));
      $this->params->set('optimizeCss', $componentParams->get('optimizeCss', 0));
      $this->params->set('minifyCss', $componentParams->get('minifyCss', 1));
      $this->params->set('optimizeCacheTime', $componentParams->get('optimizeCacheTime', 15));
      $this->params->set('optimizeExclusions', $componentParams->get('optimizeExclusions', array()));

      $option = $application->input->getCmd('option');

      if ($user->guest && $application->isSite() && $application->input->getMethod() == 'GET' && $document->getType() == 'html' && class_exists('DOMDocument') && !in_array($option, $this->params->get('optimizeExclusions', array())) && ($this->params->get('optimizeJs') || $this->params->get('optimizeCss'))) {

        $key = md5(JUri::current());

        $response = '';
        if($this->params->get('optimizeCacheTime')) {
          $cache = JFactory::getCache('plg_system_route66', 'output');
          $cache->setCaching(true);
          $cache->setLifeTime($this->params->get('optimizeCacheTime'));
          $response = $cache->get($key);
        }

        if(!$response) {

          $buffer = $application->getBody();
          $dom = new DOMDocument();
          $dom->preserveWhiteSpace = false;
          @$dom->loadHTML($buffer);

          if($this->params->get('optimizeJs')) {
            $this->optimizeJs($dom, $key);
          }

          if($this->params->get('optimizeCss')) {
            $this->optimizeCss($dom, $key);
          }

          $response = $dom->saveHTML();
          $cache->store($response, $key);

        }
        $application->setBody($response);
      }

    }

    protected function optimizeJs(DOMDocument &$dom, $key) {
      $js = '';
      $scripts = $dom->getElementsByTagName('script');
      foreach (iterator_to_array($scripts) as $script) {
          $type = $script->getAttribute('type');
          if($type == 'text/javascript' || $type == '') {
            $src = $script->getAttribute('src');
            if ($src) {
                $script->setAttribute('defer', 'defer');
            } else {
                $js .= $script->textContent;
                $script->parentNode->removeChild($script);
            }
          }
      }
      if($js) {
        $script = $dom->createElement('script');
        if($this->params->get('optimizeJsMethod') == 'file') {
          jimport('joomla.filesystem.file');
          $date = JFactory::getDate();
          JFile::write(JPATH_SITE.'/media/route66/scripts/'.$key.'.js', $js);
          $script->setAttribute('src', JUri::root(true).'/media/route66/scripts/'.$key.'.js?v='.$date->toUnix());
          $script->setAttribute('defer', 'defer');
        } else {
          $script->textContent = 'document.addEventListener("DOMContentLoaded", function(event) {'.$js.'});';
        }
        $dom->getElementsByTagName('body')->item(0)->appendChild($script);
      }
    }

    protected function optimizeCss(DOMDocument &$dom, $key) {

      $rootRelative = JUri::root(true);
      $rootAbsolute = JUri::root(false);

      $style = $dom->createElement('style');
      $children = $dom->getElementsByTagName('head')->item(0)->childNodes;

      foreach (iterator_to_array($children) as $child) {

        if(!isset($child->tagName)) {
          continue;
        }

        if($child->tagName == 'style') {

          $style->textContent .= $child->textContent;
          $child->parentNode->removeChild($child);

        } else if($child->tagName == 'link') {

          $rel = $child->getAttribute('rel');

          if ($rel == 'stylesheet') {
              $href = $child->getAttribute('href');
              if(strpos($href, '//') === 0) {
                continue;
              } else if(strpos($href, $rootAbsolute) === 0) {
                $href = str_replace($rootAbsolute, '', $href);
              } else if(strpos($href, $rootRelative) === 0) {
                $href = str_replace($rootRelative.'/', '', $href);
              } else {
                continue;
              }

              $filepath = parse_url($href, PHP_URL_PATH);
              $basepath = dirname($filepath);
              $parts = explode('/', $basepath);
              $css = @file_get_contents(JPATH_SITE.'/'.$filepath);

              preg_match_all('/url\((.*?)\)/i', $css, $matches);
              if(isset($matches[1])) {
                $replacements = array();
                foreach($matches[1] as $url) {
                  $url = trim($url, '\'"');
                  if(strpos($url, 'data:') === 0) {
                    $replacements[] = $url;
                    continue;
                  }
                  $count = substr_count($url, '../');
                  if($count) {
                    $url = str_replace(str_repeat('../', $count), implode('/', array_slice($parts, 0, -$count)).'/', $url);
                    $url = ltrim($url, '/');
                  } else if(strpos($url, './') === 0) {
                    $url = str_replace('./', $basepath);
                  } else if(strpos($url, '/') === 0) {
                  } else {
                    $url = $basepath.'/'.$url;
                  }
                  $replacements[] = $rootRelative.'/'.$url;

                }
                $css = str_replace($matches[1], $replacements, $css);
              }
              $style->textContent .= $css;
              $child->parentNode->removeChild($child);
            }
        }

      }

      if($this->params->get('minifyCss')) {
        require_once JPATH_SITE.'/plugins/system/route66/lib/minifier.php';
        $minifier = new Route66CssMinifier();
        $style->textContent = $minifier->process($style->textContent);
      }
      $dom->getElementsByTagName('head')->item(0)->appendChild($style);
    }

}
